from torch.utils.data import Dataset, DataLoader
class Cudaset(Dataset):
    def __init__(self, dataset: Dataset):
        (X, y) = next(iter(DataLoader(dataset, len(dataset))))
        self.X, self.y = X.cuda(), y.cuda()
    def __len__(self): return len(self.y)
    def __getitem__(self, index): return self.X[index], self.y[index]



import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torch.optim.lr_scheduler as sched
from torchsummary import summary
import matplotlib.pyplot as plt
from tqdm.notebook import tqdm
from .colab import ColabOutput
from .tqdn import tqdn

class Scheduler:
    def __init__(self, scheduler, optimizer, T_max):
        self.scheduler = scheduler
        self.optimizer = optimizer
        self.T_max = T_max
    def init(self, epochs):
        if self.T_max == -1: self.T_max = epochs
        if self.scheduler: self.scheduler = self.scheduler(self.optimizer, self.T_max)
    def step(self):
        if self.scheduler: self.scheduler.step()

class Module(nn.Module):
    def forward(self, input):
        return self.module(input)

    def init(self, optimizer=optim.Adam, lr=1e-3, scheduler=sched.CosineAnnealingLR, T_max=-1):
        self.optimizer = optimizer(self.module.parameters(), lr)
        self.scheduler = Scheduler(scheduler, self.optimizer, T_max)
        return self.cuda()

    def summary(self, *input_size):
        return summary(self.module.cuda(), (input_size))

    def test(self, loader):
        self.eval()
        accuracy = 0
        with torch.no_grad():
            for bar, (images, labels) in tqdn(loader, desc='Testing\t'):
                predict = self(images).argmax(1)
                accuracy += (predict == labels).sum().item()
                bar.set_postfix_str('Accuracy is {:%}'.format(accuracy / len(loader.dataset)))

    def fit(self, epochs:int, loaders:list, losser=F.cross_entropy):
        colab = ColabOutput()
        losses = [[], [], []]
        self.scheduler.init(epochs)
        for bar, epoch in tqdn(range(epochs), desc='Training'):
            for param_group in self.optimizer.param_groups:
                bar.set_postfix_str(f'lr:{param_group["lr"]}')
                break

            self.train()
            for loss in losses: loss.append(0)
            for images, labels in loaders[0]:
                self.optimizer.zero_grad()
                loss = losser(self(images), labels)
                loss.backward(), self.optimizer.step()
                losses[0][-1] += loss.item()
            self.scheduler.step()
            losses[0][-1] /= len(loaders[0])

            if len(loaders) < 2: continue
            else: self.eval()
            with torch.no_grad():
                for images, labels in loaders[1]:
                    output = self(images)
                    losses[1][-1] += losser(output, labels).item()
                    losses[2][-1] += (output.argmax(1) == labels).sum().item()
            losses[1][-1] /= len(loaders[1])
            losses[2][-1] /= len(loaders[1].dataset)

            with colab:
                plt.plot(losses[0], label='Training')
                plt.plot(losses[1], label='Validation')
                plt.plot(losses[2], label='Accuracy')
                plt.legend()
        return self



import IPython, ipywidgets, os
import matplotlib.pyplot as plt
class ColabOutput(object):
    def __init__(self, locals=None):
        if not hasattr(self, 'output'):
            from google.colab import output
            output.enable_custom_widget_manager()
            self.output = ipywidgets.Output()
            IPython.display.display(self.output)
        PingColab()
    def __enter__(self):
        self.output.__enter__()
        IPython.display.clear_output(True)
    def __exit__(self, *args):
        plt.show()
        self.output.__exit__(*args)
    def __new__(cls, locals=None):
        if not locals:
            return super().__new__(cls)
        if 'c01ab' not in locals:
            locals['c01ab'] = super().__new__(cls)
        return locals['c01ab']

class ColabPlot(ColabOutput): pass
def PingColab() -> None:
    if os.path.exists('/content/stop'): 1/0



from tqdm.notebook import tqdm
class tqdn():
    def __init__(self, iterable, **args):
        self.tqdm = tqdm(iterable, **args)
    def __iter__(self):
        for item in self.tqdm:
            yield self.tqdm, item



import torch, warnings
import torch.nn.functional as F

def _fspecial_gauss_1d(size, sigma):
    coords = torch.arange(size, dtype=torch.float)
    coords -= size // 2
    g = torch.exp(-(coords ** 2) / (2 * sigma ** 2))
    g /= g.sum()
    return g.unsqueeze(0).unsqueeze(0)

def gaussian_filter(input, win):
    assert all([ws == 1 for ws in win.shape[1:-1]]), win.shape
    if len(input.shape) == 4:
        conv = F.conv2d
    elif len(input.shape) == 5:
        conv = F.conv3d
    else:
        raise NotImplementedError(input.shape)
    C = input.shape[1]
    out = input
    for i, s in enumerate(input.shape[2:]):
        if s >= win.shape[-1]:
            out = conv(out, weight=win.transpose(2 + i, -1), stride=1, padding=0, groups=C)
        else:
            warnings.warn(f'Skipping Gaussian Smoothing at dimension 2+{i} for input: {input.shape} and win size: {win.shape[-1]}')
    return out

def _ssim(X, Y, data_range, win, size_average, K):
    K1, K2 = K
    compensation = 1.0
    C1 = (K1 * data_range) ** 2
    C2 = (K2 * data_range) ** 2
    win = win.to(X.device, dtype=X.dtype)
    mu1 = gaussian_filter(X, win)
    mu2 = gaussian_filter(Y, win)
    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2
    sigma1_sq = compensation * (gaussian_filter(X * X, win) - mu1_sq)
    sigma2_sq = compensation * (gaussian_filter(Y * Y, win) - mu2_sq)
    sigma12 = compensation * (gaussian_filter(X * Y, win) - mu1_mu2)
    cs_map = (2 * sigma12 + C2) / (sigma1_sq + sigma2_sq + C2)
    ssim_map = ((2 * mu1_mu2 + C1) / (mu1_sq + mu2_sq + C1)) * cs_map
    ssim_per_channel = torch.flatten(ssim_map, 2).mean(-1)
    cs = torch.flatten(cs_map, 2).mean(-1)
    return ssim_per_channel, cs

def ssim(X, Y, data_range=255, size_average=True, win_size=11, win_sigma=1.5, win=None, K=(0.01,0.03), nonnegative_ssim=False):
    if not X.shape == Y.shape:
        raise ValueError(f'Input images should have the same dimensions, but got {X.shape} and {Y.shape}.')
    for d in range(len(X.shape) - 1, 1, -1):
        X = X.squeeze(dim=d)
        Y = Y.squeeze(dim=d)
    if len(X.shape) not in (4, 5):
        raise ValueError(f'Input images should be 4-d or 5-d tensors, but got {X.shape}')
    if win is not None:
        win_size = win.shape[-1]
    if not (win_size % 2 == 1):
        raise ValueError('Window size should be odd.')
    if win is None:
        win = _fspecial_gauss_1d(win_size, win_sigma)
        win = win.repeat([X.shape[1]] + [1] * (len(X.shape) - 1))
    ssim_per_channel, cs = _ssim(X, Y, data_range=data_range, win=win, size_average=False, K=K)
    if nonnegative_ssim:
        ssim_per_channel = torch.relu(ssim_per_channel)
    if size_average:
        return ssim_per_channel.mean()
    else:
        return ssim_per_channel.mean(1)



from tqdm.notebook import tqdm
from duy_book.tqdn import tqdn
from duy_book.pytorch_ssim import ssim
from duy_book.ratio_split import ratio_split
from duy_book.random_split import random_split
from duy_book.colab import ColabOutput, ColabPlot, PingColab
from duy_book.cudaset import Cudaset
from duy_book.module import Module



from torch.utils.data import random_split
def ratio_split(dataset, ratios):
    #assert sum(ratios) == 1
    lengths = [int(r*len(dataset)) for r in ratios]
    if len(dataset) != sum(lengths):
        if ratios[-1] == -1:
            lengths[-1] += len(dataset) - sum(lengths)
        else:
            lengths.append(len(dataset) - sum(lengths))
    return random_split(dataset, lengths)
